class Fibonacci:
    """Class Fibonacci is used to calculate a series of Fibonacci numbers.


    Attributes
    ----------

    _cache : List[int]
        A list of already calculated Fibonacci numbers.


    Methods
    -------

    __str__()
        Returns a string representation of a class object.
    __getitem__()
        Returns item by its index.
    __extend_cache()
        Is used to extend cache list of Fibonacci numbers up to given index.


    Example
    -------

    >>> from dynalglib import Fibonacci
    >>> fibonacci = Fibonacci()
    >>> print(fibonacci[5]) # prints 5th Fibonacci number

    """

    def __init__(self) -> None:
        self._cache = [0, 1]

    def __str__(self) -> str:
        """Returns a string representation of a class object.

        Parameters
        ----------
        None

        Returns
        -------
        str
            A string representation of a class object.
        """

        return str(self._cache)

    def __getitem__(self, key) -> int:
        """Returns item by its index.

        Parameters
        ----------
        key : int
            An index of the desired element.

        Returns
        -------
        int
            A desired element.

        Raises
        ------
        ValueError
            If given parameter key is negative.
        """

        if key < 0:
            raise ValueError("Index must be non-negative")

        if key >= len(self._cache):
            self.__extend_cache(key)
        return self._cache[key]

    def __extend_cache(self, n) -> None:
        """Is used to extend cache list of Fibonacci numbers up to given index.

        Parameters
        ----------
        n : int
            An index of the element up to which the cache will be extended.

        Returns
        -------
        None
        """

        for i in range(len(self._cache), n + 1):
            self._cache.append(self._cache[i - 1] + self._cache[i - 2])



from typing import List


class Coin_Change:
    """Class Coin_Change is used to solve the Coin Change Problem.


    Attributes
    ----------

    coins : dict[]
        A dictionary in which the key is the denomination of the coin, and the value is the number of coins needed.
    denominations : List[int]
        A list of available denominations.
    count_coins: int
        A minimum possible number of used coins.


    Methods
    -------

    __str__()
        Returns a string representation of a class object.
    __getitem__()
        Returns number of used coins of specified denomination.
    solve()
        Finds the combination of coins that makes up the given amount or the closest possible amount.


    Example
    -------

    >>> from dynalglib import Coin_Change
    >>> coin_change = Coin_Change()
    >>> coin_change.solve(32)
    >>> print(coin_change.coins) # prints the result "{1: 0, 2: 1, 5: 0, 10: 3}"

    """

    def __init__(self, denominations: List[int] = [1, 2, 5, 10]) -> None:
        self.coins: dict = {}
        self.denominations: List[int] = sorted(denominations)
        self.count_coins: int = 0

    def __str__(self) -> str:
        """Returns a string representation of a class object.

        Parameters
        ----------
        None

        Returns
        -------
        str
            A string representation of a class object.
        """

        return "\t".join(f"{nominal}:{count}" for nominal, count in self.coins.items())

    def __getitem__(self, key) -> int:
        """Returns number of used coins of specified denomination.

        Parameters
        ----------
        key: int
            A specified denomination.

        Returns
        -------
        int
            A number of used coins of specified denomination.
        """

        return self.coins.get(key, 0)

    def solve(self, amount) -> None:
        """Finds the combination of coins that makes up the given amount or the closest possible amount.
        If it's not possible to make up the exact amount with the given denominations,
        the method stores the combination that makes up the closest possible amount in the 'result' attribute.

        Parameters
        ----------
        amount: int
            An amount for which to find the coin combination.

        Returns
        -------
        None
        """

        dp = [float("inf")] * (amount + 1)
        dp[0] = 0

        for i in range(1, amount + 1):
            for coin in self.denominations:
                if coin <= i and dp[i - coin] + 1 < dp[i]:
                    dp[i] = dp[i - coin] + 1

        closest_amount = amount
        while dp[closest_amount] == float("inf"):
            closest_amount -= 1

        coins_used = {coin: 0 for coin in self.denominations}
        remaining_amount = closest_amount
        for coin in reversed(self.denominations):
            while remaining_amount >= coin:
                coins_used[coin] += 1
                remaining_amount -= coin

        self.coins = coins_used
        self.count_coins = sum(list(self.coins.values()))



from typing import Union


class Item:
    """Class Item is used to store data about an item.


    Attributes
    ----------

    name : Union[str, int]
        The name of the item.
    weight : int
        The weight of the item.
    value : int
        The value of the item.
    quantity : int
        The quantity of the item.


    Methods
    -------

    __str__()
        Returns a string representation of a class object.


    Example
    -------

    >>> from dynalglib import Item
    >>> my_item = Item(name="apple", weight=2, value=3, quantity=6)
    >>> print(my_item)

    """

    def __init__(
        self, name: Union[str, int], weight: int = 1, value: int = 1, quantity: int = 1
    ) -> None:
        self.name = name
        self.weight = weight
        self.value = value
        self.quantity = quantity

    def __str__(self) -> str:
        """Returns a string representation of a class object.

        Parameters
        ----------
        None

        Returns
        -------
        str
            A string representation of a class object.
        """

        return (
            "name={}".format(self.name).ljust(20)
            + " weight={}".format(self.weight).ljust(15)
            + " value={}".format(self.value).ljust(15)
            + " quantity={}".format(self.quantity).ljust(15)
        )



from typing import List


def generate_matrix(rows: int = 5, cols: int = 5) -> List[List[int]]:
    """Returns a zero matrix (two-dimensional List) of a given size.

    Parameters
    ----------
    rows : int, optional
        A number of rows in matrix, by default 5.
    cols : int, optional
        A number of colomns in matrix, by default 5.

    Returns
    -------
    List[List[int]]
        A two-dimensional List representing a zero matrix of a given size.
    """

    matrix = [[0] * cols for _ in range(rows)]
    return matrix



from typing import List
from dynalglib.item import Item
from dynalglib.utils import generate_matrix


class Knapsack:
    """Class Knapsack is used to solve the Knapsack problem.


    Attributes
    ----------

    all_items : List[Items]
        A List of all items which are available.
    capacity : int
        A maximum capacity of the knapsack.
    chosen_items : List[Items]
        A List of the chosen items to fill the knapsack.
    total_value : int
        A resulting maximum value that we obtain when knapsack is filled.


    Methods
    -------

    __len__()
        Returns number of uniq items in the knapsack.
    __str__()
        Returns a string representation of the filled knapsack.
    __getitem__()
        Returns item from filled knapsack by its index.
    _collect_answers()
        Returns List of chosen items. This function uses result_matrix to form List of answers.
    fill()
        Fills the knapsack. This function solves the knapsack problem.


    Example
    -------

    >>> from dynalglib import Knapsack
    >>> from dynalglib import Item
    >>> item1 = Item(name="apple", weight=2, value=3, quantity=6)
    >>> item2 = Item(name="pineapple", weight=4, value=8, quantity=3)
    >>> item3 = Item(name="kiwi", weight=3, value=6, quantity=3)
    >>> knapsack = Knapsack(items=[item1, item2, item3], capacity=10)
    >>> knapsack.fill()
    >>> print(knapsack.total_value) # prints max value that we obtain
    >>> print(knapsack) # prints which items we choose

    """

    def __init__(self, items: List[Item], capacity: int) -> None:
        self.all_items = items
        self.capacity = capacity
        self.chosen_items = list()
        self.total_value = 0

    def __len__(self) -> int:
        """Returns number of uniq items in the knapsack.

        Parameters
        ----------
        None

        Returns
        -------
        int
            A number of uniq items in the knapsack.
        """

        return len(self.chosen_items)

    def __str__(self) -> str:
        """Returns a string representation of the filled knapsack.

        Parameters
        ----------
        None

        Returns
        -------
        str
            A string representation of the filled knapsack.
        """

        return "\n".join(item.__str__() for item in self.chosen_items)

    def __getitem__(self, key: int) -> Item:
        """Returns item from filled knapsack by its index.

        Parameters
        ----------
        key: int
            An index of the desired item.

        Returns
        -------
        Item
            An item from filled knapsack by its index.
        """

        return self.chosen_items[key]

    def _collect_answers(self, result_matrix: List[List[int]]) -> List[Item]:
        """Returns List of chosen items. This function uses result_matrix to form List of answers.

        Parameters
        ----------
        result_matrix : List[List[int]]
            A two-dimensional List of maximum value and quntity of each item on each iteration.

        Returns
        -------
        List[Item]
            A list of chosen items.
        """

        result_List: List = []
        taken_quantity_sum: int = 0
        for i in range(len(result_matrix) - 1, -1, -1):
            if i == len(result_matrix) - 1:
                result_List.append(result_matrix[i][len(result_matrix[i]) - 1])
                taken_quantity_sum += (
                    self.all_items[i].weight
                    * result_matrix[i][len(result_matrix[i]) - 1][1]
                )
            else:
                result_List.append(
                    result_matrix[i][(len(result_matrix[i]) - 1 - taken_quantity_sum)]
                )
                taken_quantity_sum += (
                    self.all_items[i].weight
                    * result_matrix[i][len(result_matrix[i]) - 1 - taken_quantity_sum][
                        1
                    ]
                )

        result_List.reverse()

        items_info = []
        for index_of_item, answer in enumerate(result_List):
            if answer[1] > 0:
                items_info.append(
                    (
                        self.all_items[index_of_item].name,
                        self.all_items[index_of_item].weight,
                        self.all_items[index_of_item].value,
                        answer[1],
                    )
                )
        items = [
            Item(name=name, weight=weight, value=value, quantity=quantity)
            for name, weight, value, quantity in items_info
        ]

        return items

    def fill(self) -> None:
        """Fills the knapsack. This function solves the knapsack problem.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """

        ROWS = len(self.all_items)
        COLS = self.capacity + 1
        matrix = generate_matrix(rows=ROWS, cols=COLS)
        matrix = [[[None, None] for _ in range(COLS)] for _ in range(ROWS)]
        for index_row, row in enumerate(matrix):
            a = [
                min(
                    self.all_items[index_row].quantity,
                    i // self.all_items[index_row].weight,
                )
                for i in range(COLS)
            ]
            for index_col, col in enumerate(row):
                if index_row == 0:
                    col[0] = self.all_items[index_row].value * a[index_col]
                    col[1] = a[index_col]
                else:
                    results = []
                    for i in range(index_col + 1):
                        value = (
                            self.all_items[index_row].value * a[i]
                            + matrix[index_row - 1][index_col - i][0]
                        )
                        results.append(value)
                    max_in_tmp = max(results)
                    col[0] = max_in_tmp
                    col[1] = a[results.index(max_in_tmp)]
        self.chosen_items = self._collect_answers(matrix)
        self.total_value = sum([i.value * i.quantity for i in self.chosen_items])



"""
# dynalglib

Dynalglib is a library which is designed to solve some of dynamic programming algorithms in Python.


## Released algorithms:
- Fibonacci
- Coin Change Problem
- Resource Allocation Problem
- Knapsack Problem
- Traveling Salesman Problem


## Installing
To install this library you should run:
>>> pip install dynalglib    
     
To update the library you should run
>>> pip install -U dynalglib


## A simple example
``` python
from dynalglib import Item
from dynalglib import Knapsack

knapsack = Knapsack(
    items=[
        Item(name="apple", weight=2, value=3, quantity=6),
        Item(name="pineapple", weight=4, value=8, quantity=3),
        Item(name="kiwi", weight=3, value=6, quantity=3),
        Item(name="strawberry", weight=1, value=7, quantity=8),
    ],
    capacity=10,
)

knapsack.fill() 
print(knapsack.total_value)
print(knapsack)
```
"""

from .item import Item
from .knapsack import Knapsack
from .tsp import TSP
from .coin_change import Coin_Change
from .fibonacci import Fibonacci
from .resource_allocation import Resource_Allocation
from .utils import *



from distutils.core import setup
import pathlib

VERSION = "1.1"

setup(
    name="dynalglib",  # How you named your package folder (MyLib)
    packages=["dynalglib"],  # Chose the same as "name"
    version=VERSION,  # Start with a small number and increase it with every change you make
    license="MIT",  # Chose a license from here: https://help.github.com/articles/licensing-a-repository
    description="Dynalglib is a library which is designed to solve some of dynamic programming algorithms in Python.",  # Give a short description about your library
    long_description=pathlib.Path("README.md").read_text(),
    long_description_content_type="text/markdown",
    author="Milana Antonova",  # Type in your name
    author_email="mmf.antonovam@gmail.com",  # Type in your E-Mail
    url="https://github.com/yslsml/dynalglib",  # Provide either the link to your github or to your website
    project_urls={
        "Documentation":"https://github.com/yslsml/dynalglib/wiki/Documentation",
        "Source": "https://github.com/yslsml/dynalglib"
    },
    keywords=[
        "math",
        "dynamic algo",
        "algorithms",
        "algorithm",
        "dynamic programming",
        "knapsack problem",
    ],  # Keywords that define your package best
    classifiers=[
        "Development Status :: 3 - Alpha",  # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        "Intended Audience :: Developers",  # Define that your audience are developers
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",  # Again, pick a license
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
)



import random
from typing import List
from dynalglib.item import Item
from dynalglib.utils import generate_matrix


class Resource_Allocation:
    """Class Resource_Allocation is used to solve the Resource Allocation problem.


    Attributes
    ----------

    profit_matrix : List[List[int]]
        A matrix with profits of each company.
    invest_in_company : List[int]
        A list representing the amount we allocate to each company.
    max_profit : int
        A maximum profit that can be obtained.


    Methods
    -------

    __getitem__()
        Returns the amount that we allocate in company by its index.
    collect_answers()
        Returns list representing the amount we allocate to each company.
    solve()
        Solves the Resource Allocation problem.
    generate_random_profit_matrix()
        Generates random profit matrix.


    Example
    -------

    >>> from dynalglib import Resource_Allocation
    >>> profit_matrix = Resource_Allocation.generate_random_profit_matrix(5,10,1,50)
    >>> resource_allocation = Resource_Allocation(profit_matrix)
    >>> resource_allocation.solve()
    >>> print(resource_allocation.max_profit) # prints max profit that we obtain
    >>> print(resource_allocation.invest_in_company) # prints the allocation

    """

    def __init__(self, profit_matrix: List[List[int]]):
        self.profit_matrix = [[0] + row for row in profit_matrix]
        self.invest_in_company: List[int] = list()
        self.max_profit: int = 0

    def __getitem__(self, key: int) -> int:
        """Returns the amount that we allocate in company by its index.

        Parameters
        ----------
        key: int
            An index of the desired company.

        Returns
        -------
        int
            The amount that we allocate in desired company.
        """

        return self.invest_in_company[key]

    def _collect_answers(self, result_matrix: List[List[int]]) -> List[int]:
        """Returns list representing the amount we allocate to each company.
        This function uses result_matrix to form allocation list.

        Parameters
        ----------
        result_matrix : List[List[int]]
            A two-dimensional list of maximum profit and amount that could be allocated to each company on each iteration.

        Returns
        -------
        List[int]
            A list representing the amount we allocate to each company.
        """

        results: List = []
        taken_quantity_sum: int = 0

        for i in range(len(result_matrix) - 1, -1, -1):
            if i == len(result_matrix) - 1:
                results.append(result_matrix[i][len(result_matrix[i]) - 1])
                taken_quantity_sum += result_matrix[i][len(result_matrix[i]) - 1][1]
            else:
                results.append(
                    result_matrix[i][(len(result_matrix[i]) - 1 - taken_quantity_sum)]
                )
                taken_quantity_sum += result_matrix[i][
                    len(result_matrix[i]) - 1 - taken_quantity_sum
                ][1]

        results.reverse()
        return results

    def solve(self) -> None:
        """Solves Resource Allocation problem.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """

        ROWS = len(self.profit_matrix)
        COLS = len(self.profit_matrix[0])
        matrix = generate_matrix(rows=ROWS, cols=COLS)
        matrix = [[[None, None] for _ in range(COLS)] for _ in range(ROWS)]
        for index_row, row in enumerate(matrix):
            for index_col, col in enumerate(row):
                if index_row != 0:
                    results = []
                    for i in range(index_col + 1):
                        value = (
                            self.profit_matrix[index_row][i]
                            + matrix[index_row - 1][index_col - i][0]
                        )
                        results.append(value)
                    max_value = max(results)
                    col[0] = max_value
                    col[1] = results.index(max_value)
                else:
                    col[0] = self.profit_matrix[index_row][index_col]
                    col[1] = index_col
        answers = self._collect_answers(result_matrix=matrix)
        self.invest_in_company = [i[1] for i in answers]
        self.max_profit = answers[-1][0]

    @staticmethod
    def generate_random_profit_matrix(
        company_count: int = 5,
        max_investment: int = 10,
        min_profit: int = 1,
        max_profit: int = 12,
    ) -> List[List[int]]:
        """Generates random profit matrix.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """

        matrix = generate_matrix(rows=company_count, cols=max_investment + 1)
        for i in range(company_count):
            for j in range(max_investment + 1):
                while True:
                    number = random.randint(min_profit, max_profit) if j != 0 else 0
                    if number == 0 or number not in matrix[i]:
                        matrix[i][j] = number
                        break
            matrix[i].sort()
        return matrix



import random
from typing import List
from dynalglib.utils import generate_matrix


class TSP:
    """Class TSP is used to solve the Traveling Salesman Problem.


    Attributes
    ----------

    distance_matrix : List[List[int]]
        A two-dimensional List (symmetric matrix) of distances between cities.
    min_distance : int
        The minimum distance for which all cities can be visited.
    path : List[int]
        A List of cities which represent the shortest possible path.


    Methods
    -------

    __str__()
        Returns a string representation of the minimum distance and resulting path.
    _dfs()
        Depth-First Search algorithm to find the shortest path in the TSP.
    solve()
        Solves the Traveling Salesman Problem.
    generate_random_distance_matrix()
        Returns two-dimensional List of random distances between cities.


    Example
    -------

    >>> from dynalglib import TSP
    >>> distance_matrix = TSP.generate_random_distance_matrix(city_count=5)
    >>> tsp = TSP(distance_matrix=distance_matrix)
    >>> tsp.solve()
    >>> print(tsp.min_distance) # prints minimum possible distance
    >>> print(tsp.path) # prints the best path

    """

    def __init__(self, distance_matrix: List[List[int]]) -> None:
        self.distance_matrix: List[List[int]] = distance_matrix
        self.min_distance: int = None
        self.path: List[int] = None

    def __str__(self) -> str:
        """Returns a string representation of the minimum distance and resulting path.

        Parameters
        ----------
        None

        Returns
        -------
        str
            A string representation of the minimum distance and resulting path.
        """

        return f"min distance: {self.min_distance}\tpath: {self.path}"

    def _dfs(self, node, visited, path) -> tuple[int, List[int]]:
        """Depth-First Search algorithm to find the shortest path in the TSP.

        Parameters
        ----------
        node : int
            The current city being visited.
        visited : List[bool]
            A List indicating whether a city has been visited or not.
        path : List[int]
            A List representing the current path traversed.

        Returns
        -------
        tuple: [int, List[int]]
            The minimum distance found and the corresponding shortest path.
        """

        visited[node] = True
        path.append(node)

        if len(path) == len(self.distance_matrix):
            return self.distance_matrix[node][path[0]], path

        min_distance = float("inf")
        min_path = None
        for i in range(len(self.distance_matrix)):
            if not visited[i]:
                distance, new_path = self._dfs(i, visited.copy(), path.copy())
                distance += self.distance_matrix[node][i]
                if distance < min_distance:
                    min_distance = distance
                    min_path = new_path

        return min_distance, min_path

    def solve(self, start: int = 0) -> None:
        """Solves the TSP.

        Parameters
        ----------
        start : int, optional
            The index of the starting city, by default 0.

        Returns
        -------
        None
        """

        n = len(self.distance_matrix)
        visited = [False] * n
        path = []

        distance, min_path = self._dfs(start, visited, path)
        min_path.append(start)
        self.min_distance = distance
        self.path = min_path

    @staticmethod
    def generate_random_distance_matrix(
        city_count: int = 5, min_distance: int = 0, max_distance: int = 10
    ) -> List[List[int]]:
        """Returns a random distance matrix between cities.

        Parameters
        ----------
        city_count : int, optional
            The number of cities, by default 5.
        min_distance : int, optional
            The minimum distance between cities, by default 0.
        max_distance : int, optional
            The maximum distance between cities, by default 10.

        Returns
        -------
        List[List[int]]
            A two-dimensional List representing the distance matrix.
        """

        matrix = generate_matrix(rows=city_count, cols=city_count)

        for i in range(city_count):
            for j in range(i + 1, city_count):
                distance = random.randint(min_distance, max_distance)
                matrix[i][j] = distance
                matrix[j][i] = distance
        return matrix
